#include <stdio.h>
#include <libojmixer/ojconfig.h>
#include "ojmixerwin.h"
#include "disclosure.h"

static void load_presets(OjMixerWindow *win);

static gint
delete_win_cb(GtkWidget *w, GdkEventAny *e, OjMixerWindow *win)
{
	oj_mixer_window_destroy(win);

	gtk_main_quit();

	return TRUE;
}

static void
mixer_vol_updated_cb(OjMixer *mixer, guint device, guint volume,
					 OjMixerWindow *win)
{
	if (gtk_range_get_value(GTK_RANGE(win->scales[device])) == volume)
		return;

	gtk_range_set_value(GTK_RANGE(win->scales[device]), volume);
}

static void
volume_changed_cb(GtkRange *range, int device)
{
	OjMixer *mixer;
	OjMixerWindow *win;

	mixer = g_object_get_data(G_OBJECT(range), "mixer");
	win   = g_object_get_data(G_OBJECT(range), "mixer_window");

	oj_mixer_set_volume(mixer, device, gtk_range_get_value(range));

	if (win->active_preset != NULL)
	{
		gtk_widget_set_sensitive(win->revert_button, TRUE);
		gtk_widget_set_sensitive(win->save_button,   TRUE);
	}
}

static void
fade_toggle_cb(GtkToggleButton *button, OjMixerWindow *win)
{
	oj_mixer_set_fade_on_switch(win->mixer,
								gtk_toggle_button_get_active(button));
}

static void
rec_switched_cb(GtkWidget *button, int device)
{
	OjMixer *mixer;
	OjMixerWindow *win;
	GtkWidget *image, *old_image;

	mixer = g_object_get_data(G_OBJECT(button), "mixer");
	win   = g_object_get_data(G_OBJECT(button), "mixer_window");

	/* Set the old image back to off */
	if (win->rec_buttons[oj_mixer_get_rec_source(mixer)] != NULL) {
		old_image =
			GTK_BIN(win->rec_buttons[oj_mixer_get_rec_source(mixer)])->child;

		gtk_image_set_from_stock(GTK_IMAGE(old_image), OJ_STOCK_RECORDING_OFF,
								 GTK_ICON_SIZE_MENU);
	}

	/* Set the new image to on */
	image = GTK_BIN(button)->child;

	gtk_image_set_from_stock(GTK_IMAGE(image), OJ_STOCK_RECORDING_ON,
							 GTK_ICON_SIZE_MENU);

	oj_mixer_set_rec_source(mixer, device);

	if (win->revert_button != NULL && win->active_preset != NULL)
		gtk_widget_set_sensitive(win->revert_button, TRUE);

	if (win->save_button != NULL && win->active_preset != NULL)
		gtk_widget_set_sensitive(win->save_button, TRUE);
}

static void
revert_preset_cb(GtkWidget *button, OjMixerWindow *win)
{
	int i;

	if (win->active_preset == NULL)
		return;

	for (i = 0; i < SOUND_MIXER_NRDEVICES; i++)
	{
		if (oj_mixer_is_device_supported(win->mixer, i))
		{
			int volume;

			volume = oj_preset_get_volume(win->active_preset, i);

			oj_mixer_set_volume(win->mixer, i, volume);

			gtk_range_set_value(GTK_RANGE(win->scales[i]), volume);
		}
	}

	oj_mixer_set_rec_source(win->mixer,
							oj_preset_get_rec_device(win->active_preset));

	gtk_widget_set_sensitive(win->revert_button, FALSE);
	gtk_widget_set_sensitive(win->save_button,   FALSE);
}

static void
save_preset_cb(GtkWidget *button, OjMixerWindow *win)
{
	if (win->active_preset == NULL)
		return;

	oj_preset_load_volume(win->active_preset, win->mixer);
	oj_preset_load_rec_device(win->active_preset, win->mixer);

	gtk_widget_set_sensitive(win->revert_button, FALSE);
	gtk_widget_set_sensitive(win->save_button,   FALSE);

	oj_config_save(win->mixer);
}

static void
preset_name_changed_cb(GtkEntry *entry, GtkDialog *dialog)
{
	gtk_dialog_set_response_sensitive(dialog, GTK_RESPONSE_ACCEPT,
									  *gtk_entry_get_text(entry) != '\0');
}

static void
new_preset_cb(GtkWidget *button, OjMixerWindow *win)
{
	OjMixer *mixer;
	OjPreset *preset;
	GtkWidget *dialog;
	GtkWidget *label;
	GtkWidget *hbox;
	GtkWidget *vbox;
	GtkWidget *entry;
	GtkWidget *image;
	gint result;

	mixer = win->mixer;

	dialog = gtk_dialog_new_with_buttons(NULL,
										 GTK_WINDOW(win->window),
										 GTK_DIALOG_MODAL,
										 GTK_STOCK_CANCEL,
										 GTK_RESPONSE_REJECT,
										 GTK_STOCK_OK,
										 GTK_RESPONSE_ACCEPT,
										 NULL);

	gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_ACCEPT);
	gtk_container_set_border_width(GTK_CONTAINER(dialog), 6);
	gtk_dialog_set_has_separator(GTK_DIALOG(dialog), FALSE);
	gtk_window_set_resizable(GTK_WINDOW(dialog), FALSE);

	gtk_dialog_set_response_sensitive(GTK_DIALOG(dialog),
									  GTK_RESPONSE_ACCEPT, FALSE);

	hbox = gtk_hbox_new(FALSE, 12);
	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), hbox);
	gtk_container_set_border_width(GTK_CONTAINER(hbox), 6);
	gtk_widget_show(hbox);

	image = gtk_image_new_from_stock(GTK_STOCK_DIALOG_QUESTION,
									 GTK_ICON_SIZE_DIALOG);
	gtk_misc_set_alignment(GTK_MISC(image), 0.50, 0.00);
	gtk_box_pack_start(GTK_BOX(hbox), image, FALSE, FALSE, 0);
	gtk_widget_show(image);

	vbox = gtk_vbox_new(FALSE, 12);
	gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 0);
	gtk_widget_show(vbox);

	label = gtk_label_new(NULL);
	gtk_label_set_markup(GTK_LABEL(label),
		_("<span weight=\"bold\" size=\"larger\">Enter a preset name</span>\n"
		  "\n"
		  "Specify the name that you wish to appear on the preset button. "
		  "Note that for now, this name cannot be changed."));
	gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
	gtk_misc_set_alignment(GTK_MISC(label), 0.50, 0.00);
	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
	gtk_widget_show(label);

	entry = gtk_entry_new();
	gtk_box_pack_start(GTK_BOX(vbox), entry, FALSE, FALSE, 0);
	gtk_entry_set_activates_default(GTK_ENTRY(entry), TRUE);
	gtk_widget_show(entry);

	g_signal_connect(G_OBJECT(entry), "changed",
					 G_CALLBACK(preset_name_changed_cb), dialog);

	result = gtk_dialog_run(GTK_DIALOG(dialog));

	if (result == GTK_RESPONSE_ACCEPT)
	{
		GtkWidget *button;

		preset = oj_preset_new();

		oj_preset_set_name(preset, gtk_entry_get_text(GTK_ENTRY(entry)));
		oj_preset_load_volume(preset, mixer);
		oj_preset_load_rec_device(preset, mixer);

		oj_mixer_add_preset(mixer, preset);
		button = oj_mixer_window_add_preset(win, preset);

		win->active_preset = preset;
		save_preset_cb(NULL, win);

		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), TRUE);

		oj_config_save(win->mixer);
	}

	gtk_widget_destroy(dialog);
}

static void
delete_preset_cb(GtkWidget *button, OjMixerWindow *win)
{
	GList *l;
	OjPreset *preset = win->active_preset;
	int index;
	GtkWidget *parent;

	l = g_list_find(win->mixer->presets, preset);
	index = g_list_index(win->mixer->presets, preset);

	if (l != NULL)
	{
		if (l->next != NULL)
			win->active_preset = l->next->data;
		else if (l->prev != NULL)
			win->active_preset = l->prev->data;
		else
			win->active_preset = NULL;
	}

	win->active_preset = NULL;

	oj_mixer_remove_preset(win->mixer, preset);


	/* Rebuild the table. */
	parent = gtk_widget_get_parent(win->presets_table);

	gtk_widget_destroy(win->presets_table);

	g_slist_free(win->preset_buttons);
	win->preset_buttons = NULL;
	win->presets_x = 0;
	win->presets_y = 0;
	load_presets(win);

	gtk_box_pack_start(GTK_BOX(parent), win->presets_table, TRUE, TRUE, 0);
	gtk_widget_show(win->presets_table);


	/* Save configuration. */
	oj_config_save(win->mixer);
}

static void
switch_preset_cb(GtkWidget *button, OjMixerWindow *win)
{
	OjPreset *preset = g_object_get_data(G_OBJECT(button), "preset");
	int device;

	win->active_preset = preset;

	device = oj_preset_get_rec_device(win->active_preset);

	oj_mixer_switch_preset(win->mixer, preset,
						   oj_mixer_get_fade_on_switch(win->mixer));
	oj_mixer_set_rec_source(win->mixer, device);

	if (device >= 0 && device < SOUND_MIXER_NRDEVICES &&
		win->rec_buttons[device] != NULL)
	{
		gtk_toggle_button_set_active(
			GTK_TOGGLE_BUTTON(win->rec_buttons[device]), TRUE);
	}

	gtk_widget_set_sensitive(win->save_button,   FALSE);
	gtk_widget_set_sensitive(win->revert_button, FALSE);
	gtk_widget_set_sensitive(win->delete_button, TRUE);
}

static void
load_presets(OjMixerWindow *win)
{
	OjMixer *mixer = win->mixer;

	win->presets_table = gtk_table_new(1, 4, FALSE);
	win->presets_table_width = 4;
	win->presets_table_height = 1;

	gtk_table_set_col_spacings(GTK_TABLE(win->presets_table), 6);
	gtk_table_set_row_spacings(GTK_TABLE(win->presets_table), 6);
	gtk_table_set_homogeneous(GTK_TABLE(win->presets_table), TRUE);

	if (mixer->presets == NULL)
	{
		win->no_presets_label = gtk_label_new(NULL);

		gtk_label_set_markup(GTK_LABEL(win->no_presets_label),
							 _("<b><i>No presets loaded</i></b>"));

		gtk_table_attach(GTK_TABLE(win->presets_table), win->no_presets_label,
						 0, win->presets_table_width, 0, 1,
						 GTK_FILL, 0, 0, 0);

		gtk_widget_show(win->no_presets_label);
	}
	else
	{
		GList *l;

		for (l = mixer->presets; l != NULL; l = l->next)
			oj_mixer_window_add_preset(win, l->data);
	}
}

static void
load_mixer(OjMixerWindow *win)
{
	GtkWidget *label;
	GtkWidget *scale;
	GtkWidget *rec_button;
	GtkWidget *first_rec_button = NULL;

	OjMixer *mixer;
	int i, y = 0;

	mixer = win->mixer;

	win->mixer_table = gtk_table_new(oj_mixer_get_device_count(mixer), 3,
									 FALSE);

	gtk_table_set_col_spacings(GTK_TABLE(win->mixer_table), 6);

	for (i = 0; i < SOUND_MIXER_NRDEVICES; i++)
	{
		if (oj_mixer_is_device_supported(mixer, i))
		{
			int volume;
			char buf[256];

			if (oj_mixer_get_volume(mixer, i, NULL, NULL, &volume) != 0)
				continue; /* FIXME - Display an error? */

			/* Create the label. */
			g_snprintf(buf, sizeof(buf), "%s:",
					   oj_mixer_get_device_label(mixer, i));

			label = gtk_label_new(buf);
			gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0);

			gtk_table_attach(GTK_TABLE(win->mixer_table), label,
							 0, 1, y, y + 1, GTK_FILL, 0, 0, 0);

			gtk_widget_show(label);

			/* Create the scale widget. */
			scale = win->scales[i] = gtk_hscale_new_with_range(0, 100, 5);
			gtk_scale_set_digits(GTK_SCALE(scale), 0);
			gtk_scale_set_draw_value(GTK_SCALE(scale), TRUE);
			gtk_scale_set_value_pos(GTK_SCALE(scale), GTK_POS_RIGHT);
			gtk_range_set_increments(GTK_RANGE(scale), 5, 10);

			gtk_table_attach(GTK_TABLE(win->mixer_table), scale,
							 1, 2, y, y + 1,
							 GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);

			gtk_widget_show(scale);

			g_object_set_data(G_OBJECT(scale), "mixer", mixer);
			g_object_set_data(G_OBJECT(scale), "mixer_window", win);

			/* Set the value on the scale. */
			gtk_range_set_value(GTK_RANGE(scale), volume);

			g_signal_connect(G_OBJECT(scale), "value_changed",
							 G_CALLBACK(volume_changed_cb), (gpointer)i);

			if (oj_mixer_can_device_record(mixer, i))
			{
				GtkWidget *image;

				/* Create the recording button. */
				rec_button = gtk_radio_button_new_from_widget(
							GTK_RADIO_BUTTON(first_rec_button));

				image = gtk_image_new_from_stock(OJ_STOCK_RECORDING_OFF,
												 GTK_ICON_SIZE_MENU);

				gtk_container_add(GTK_CONTAINER(rec_button), image);
				gtk_widget_show(image);

				win->rec_buttons[i] = rec_button;
				gtk_toggle_button_set_mode(GTK_TOGGLE_BUTTON(rec_button),
										   FALSE);

				g_object_set_data(G_OBJECT(rec_button), "mixer", mixer);
				g_object_set_data(G_OBJECT(rec_button), "mixer_window", win);

				g_signal_connect(G_OBJECT(rec_button), "clicked",
								 G_CALLBACK(rec_switched_cb), (gpointer)i);

				gtk_table_attach(GTK_TABLE(win->mixer_table), rec_button,
								 2, 3, y, y + 1, GTK_FILL, 0, 0, 0);

				gtk_widget_show(rec_button);

				if (first_rec_button == NULL)
					first_rec_button = rec_button;
			}

			y++;
		}
	}

	gtk_toggle_button_set_active(
			GTK_TOGGLE_BUTTON(win->rec_buttons[oj_mixer_get_rec_source(mixer)]),
			TRUE);

}

void
oj_mixer_window_destroy(OjMixerWindow *win)
{
	if (win == NULL)
		return;

	g_object_unref(G_OBJECT(win->mixer));

	gtk_widget_destroy(win->window);

	g_slist_free(win->preset_buttons);

	g_free(win);
}

OjMixerWindow *
oj_mixer_window_new(OjMixer *mixer)
{
	OjMixerWindow *win;
	GtkWidget *vbox1;
	GtkWidget *vbox2;
	GtkWidget *vbox3;
	GtkWidget *bbox;
	GtkWidget *disclosure;
	GtkWidget *button;
	GtkWidget *sep;

	win = g_new0(OjMixerWindow, 1);
	win->mixer = mixer;

	g_signal_connect(G_OBJECT(win->mixer), "volume_changed",
					 G_CALLBACK(mixer_vol_updated_cb), win);

	/* Create the window */
	win->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_role(GTK_WINDOW(win->window), "ojmixer");
	gtk_window_set_title(GTK_WINDOW(win->window), _("OJ Mixer"));
	gtk_window_set_resizable(GTK_WINDOW(win->window), FALSE);
	gtk_widget_realize(win->window);

	g_signal_connect(G_OBJECT(win->window), "delete_event",
					 G_CALLBACK(delete_win_cb), win);

	/* Create the parent vbox for everything. */
	vbox1 = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(win->window), vbox1);
	gtk_widget_show(vbox1);

	/* Create the vbox for the top half of the window. */
	vbox2 = gtk_vbox_new(FALSE, 12);
	gtk_container_set_border_width(GTK_CONTAINER(vbox2), 12);
	gtk_box_pack_start(GTK_BOX(vbox1), vbox2, TRUE, TRUE, 0);
	gtk_widget_show(vbox2);

	/* Create the presets table's parents. */
	win->presets_parent = gtk_vbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox2), win->presets_parent, TRUE, TRUE, 0);
	gtk_widget_show(win->presets_parent);

	/* Create the presets table. */
	load_presets(win);
	gtk_box_pack_start(GTK_BOX(win->presets_parent), win->presets_table,
					   TRUE, TRUE, 0);
	gtk_widget_show(win->presets_table);

	/* Create a separator. */
	sep = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(vbox2), sep, FALSE, FALSE, 0);
	gtk_widget_show(sep);

	/* Create the "Fade on switch" checkbox. */
	win->fade_checkbox = gtk_check_button_new_with_mnemonic(
			_("_Fade on switch"));
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(win->fade_checkbox),
								 oj_mixer_get_fade_on_switch(win->mixer));
	gtk_box_pack_start(GTK_BOX(vbox2), win->fade_checkbox, FALSE, FALSE, 0);
	gtk_widget_show(win->fade_checkbox);

	g_signal_connect(G_OBJECT(win->fade_checkbox), "toggled",
					 G_CALLBACK(fade_toggle_cb), win);

	/* Create the "Show mixer settings" disclosure. */
	disclosure = oj_disclosure_new(_("Show mixer settings"),
								   _("Hide mixer settings"));
	gtk_box_pack_start(GTK_BOX(vbox2), disclosure, FALSE, FALSE, 0);
	gtk_widget_show(disclosure);

	/* Create the second vbox, containing the mixer settings. */
	vbox3 = gtk_vbox_new(FALSE, 12);
	gtk_box_pack_start(GTK_BOX(vbox2), vbox3, TRUE, TRUE, 0);

	/* Create a separator. */
	sep = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(vbox3), sep, FALSE, FALSE, 0);
	gtk_widget_show(sep);

	/* Create the mixer table. */
	load_mixer(win);
	gtk_box_pack_start(GTK_BOX(vbox3), win->mixer_table, TRUE, TRUE, 0);
	gtk_widget_show(win->mixer_table);

	/* Create a separator. */
	sep = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(vbox3), sep, FALSE, FALSE, 0);
	gtk_widget_show(sep);

	/* Create the button box. */
	bbox = gtk_hbutton_box_new();
	gtk_box_set_spacing(GTK_BOX(bbox), 6);
	gtk_button_box_set_layout(GTK_BUTTON_BOX(bbox), GTK_BUTTONBOX_END);
	gtk_box_pack_end(GTK_BOX(vbox3), bbox, FALSE, TRUE, 0);
	gtk_widget_show(bbox);

	/* Delete button */
	win->delete_button = button = gtk_button_new_from_stock(GTK_STOCK_DELETE);
	gtk_widget_set_sensitive(button, FALSE);
	gtk_box_pack_start(GTK_BOX(bbox), button, FALSE, FALSE, 0);
	gtk_widget_show(button);

	g_signal_connect(G_OBJECT(button), "clicked",
					 G_CALLBACK(delete_preset_cb), win);

	/* New button */
	win->new_button = button = gtk_button_new_from_stock(GTK_STOCK_NEW);
	gtk_box_pack_start(GTK_BOX(bbox), button, FALSE, FALSE, 0);
	gtk_widget_show(button);

	g_signal_connect(G_OBJECT(button), "clicked",
					 G_CALLBACK(new_preset_cb), win);

	/* Defaults button */
	win->new_button = button = gtk_button_new_with_label(_("Defaults"));
	gtk_box_pack_start(GTK_BOX(bbox), button, FALSE, FALSE, 0);
	gtk_widget_show(button);

	/* Revert button */
	win->revert_button = button =
		gtk_button_new_from_stock(GTK_STOCK_REVERT_TO_SAVED);
	gtk_widget_set_sensitive(button, FALSE);
	gtk_box_pack_start(GTK_BOX(bbox), button, FALSE, FALSE, 0);
	gtk_widget_show(button);

	g_signal_connect(G_OBJECT(button), "clicked",
					 G_CALLBACK(revert_preset_cb), win);

	/* Save button */
	win->save_button = button = gtk_button_new_from_stock(GTK_STOCK_SAVE);
	gtk_widget_set_sensitive(button, FALSE);
	gtk_box_pack_start(GTK_BOX(bbox), button, FALSE, FALSE, 0);
	gtk_widget_show(button);

	g_signal_connect(G_OBJECT(button), "clicked",
					 G_CALLBACK(save_preset_cb), win);

	/* Setup the disclosure for the second vbox */
	oj_disclosure_set_container(OJ_DISCLOSURE(disclosure), vbox3);

	return win;
}

GtkWidget *
oj_mixer_window_add_preset(OjMixerWindow *win, OjPreset *preset)
{
	GtkWidget *button;

	if (win->no_presets_label != NULL)
	{
		/* Get rid of the label we have stuck in there. */
		gtk_container_remove(GTK_CONTAINER(win->presets_table),
							 win->no_presets_label);

		win->no_presets_label = NULL;
	}

	if (win->preset_buttons == NULL)
	{
		/* UGHHH gross hack so that no toggle button is showing! :) */
		win->hidden_preset = button = gtk_radio_button_new(NULL);
		gtk_toggle_button_set_mode(GTK_TOGGLE_BUTTON(button), TRUE);
		win->preset_buttons = g_slist_append(win->preset_buttons, button);
		g_object_ref(button);
	}

	button = gtk_radio_button_new_with_label(win->preset_buttons,
											 oj_preset_get_name(preset));
	win->preset_buttons = g_slist_append(win->preset_buttons, button);

	gtk_toggle_button_set_mode(GTK_TOGGLE_BUTTON(button), FALSE);
	gtk_toggle_button_set_inconsistent(GTK_TOGGLE_BUTTON(button), FALSE);

	g_object_set_data(G_OBJECT(button), "preset", preset);

	g_signal_connect(G_OBJECT(button), "clicked",
					 G_CALLBACK(switch_preset_cb), win);

	if (win->presets_x == win->presets_table_width)
	{
		win->presets_x = 0;
		win->presets_y++;

		if (win->presets_y >= win->presets_table_height)
		{
			gtk_table_resize(GTK_TABLE(win->presets_table),
							 win->presets_table_height + 1,
							 win->presets_table_width);

			win->presets_table_height++;
		}
	}

	gtk_table_attach(GTK_TABLE(win->presets_table), button,
					 win->presets_x, win->presets_x + 1,
					 win->presets_y, win->presets_y + 1,
					 GTK_FILL, GTK_FILL, 0, 0);
	gtk_widget_show(button);

	win->presets_x++;

	return button;
}
